import bpy
import logging
from . import NodeSocket

class HiveNode:
  bl_icon = 'NODETREE'
  bl_generic_sockets = True 
    
  # Copy function to initialize a copied node from an existing one.
  def copy(self, node):
    raise NotImplementedError    
  
  def set_attributes(self, attributes):
    for anr,a in enumerate(attributes):
      assert a.inhook or a.outhook #TODO?      
      if a.outhook:
        h = a.outhook
        if a.inhook: 
          socktype = NodeSocket.socketclasses[h.shape, h.color, True] #complementary socket
        else:
          socktype = NodeSocket.socketclasses[h.shape, h.color]
        sock = self.outputs.new(socktype.bl_idname, a.name)
        sock.link_limit = 99
        sock.row = anr + 1        
      if a.inhook:
        h = a.inhook
        socktype = NodeSocket.socketclasses[h.shape, h.color]
        sock = self.inputs.new(socktype.bl_idname, a.name)
        sock.link_limit = 99
        sock.row = anr + 1
      
  def check_update(self):
    for s in self.inputs: s.check_update()
    for s in self.outputs: s.check_update()
  
  def find_input_socket(self, name):
    for s in self.inputs: 
      if s.name == name: return s
    return AttributeError(name)  
    
  def find_output_socket(self, name):
    for s in self.outputs: 
      if s.name == name: return s
    return AttributeError(name)  
    
  def draw_buttons(self, arg1, arg2):
    nodetree = self.id_data
    if nodetree.nodes[0].label == self.label:
      nodetree.full_update()

  def draw_buttons_ext(self, context, layout):
    from . import BlendManager
    nodetree = self.id_data
    bntm = BlendManager.blendmanager.get_nodetree_manager(nodetree.name)
    bntm.mainWin.h().draw_panel(context, layout)
    
class HivemapNode(bpy.types.Node, HiveNode):
  bl_idname = "HivemapNode"
  bl_label = bl_idname
  bl_generic_sockets = True
  bl_width_default = 50
  @classmethod
  def poll(cls, ntree):
    return ntree.bl_idname == 'Hivemap'

class WorkermapNode(bpy.types.Node, HiveNode):
  bl_idname = "WorkermapNode"
  bl_label = bl_idname
  @classmethod
  def poll(cls, ntree):
    return ntree.bl_idname == 'Workermap'
    
class SpydermapNode(bpy.types.Node, HiveNode):
  bl_idname = "SpydermapNode"
  bl_label = bl_idname
  @classmethod
  def poll(cls, ntree):
    return ntree.bl_idname == 'Spydermap'
    
def register():
  bpy.utils.register_class(HivemapNode)
  bpy.utils.register_class(WorkermapNode)
  bpy.utils.register_class(SpydermapNode)

def unregister():
  bpy.utils.unregister_class(HivemapNode)
  bpy.utils.unregister_class(WorkermapNode)
  bpy.utils.unregister_class(SpydermapNode)
  