import logging

import spyder.formtools
from spyder.formtools import model, controller, comparator, mvc
import functools

from ..params import typemap
from .blenderview import blenderview, reserved

def _PButtonCallback(callback, model, contains_reserved):
  val = model.get()
  if val is not None and contains_reserved:
    val = val.copy()
    for k in reserved:
      kk = k + "_"
      if kk in val:
        val[k] = val[kk]
        del val[kk]
  callback(val)

def remap_reserved(callback, params):
  params2 = params.copy()
  for k in reserved:
    kk = k+"_"
    if kk in params2:
      params2[k] = params2[kk]
      del params2[kk]
  callback(params2)

def PGenerator(paramnames, paramtypelist, paramvalues, 
  update_callback, buttons = []
 ):
  contains_reserved = False
  paramnames2 = paramnames

  for k in reserved:
    if k in paramnames:
      contains_reserved = True
      paramnames2 = [p if p != k else k + "_" for p in paramnames2]

  params = list(zip(paramnames2, paramtypelist))
  typetree = spyder.formtools.generate_typetree(params,typemap)
  form = spyder.core.spyderform(typetree)  
  for subform in form._members.values():
    if subform.arraycount:
      subform.length = 10
  for buttonname, buttoncallback in buttons:    
    form.add_button(buttonname, "after")
       
  m = model(typetree)
  v = blenderview(form)
  con = controller(form, m)  
  com = comparator(form)  
  
  #link model and comparator
  m.listen(com._set)  

  #construct mvc
  my_mvc = mvc(form, lambda:m, lambda: v, lambda:con, lambda:com)

  #bind mvc
  my_mvc.bind()

  #set the current values into the model
  for pname,pvalue in paramvalues.items():
    #print("PARAMV", pname, pvalue)
    pname2 = pname
    if pname in reserved: pname2 = pname + "_"
    if pvalue is not None:
      getattr(con,pname2)._set(pvalue)

  my_mvc.sync_m2v()

  for buttonindex,button in enumerate(buttons):
    buttonname, buttoncallback = button
    cb = functools.partial(_PButtonCallback, buttoncallback, m, contains_reserved)
    v.buttons[buttonindex].listen(cb)

  up = update_callback
  if contains_reserved: up = functools.partial(remap_reserved, update_callback)
  m.listen(up)
  
  return v._widget, my_mvc