import bpy
from bpy_types import Node

class HiveNodeProxy:
  bl_icon = 'NODETREE'
  bl_generic_sockets = True 
  
  def init(self):
    x,y = self.location
    self.nodecanvas.clipboard.set_dragboard_value(self.nodetype, self.nodename)
    self.nodecanvas.drop_worker(x, y)
  
  @classmethod
  def poll(cls, ntree):
    return ntree.bl_idname == self.nodetreetype

nodecounter = 0    
class NodeManager:
  def __init__(self, clipboard, nodetree, nodetreetype):
    assert nodetreetype in ("Hivemap", "Workermap", "Spydermap"), nodetreetype    
    self.nodetree = nodetree
    self.nodetreetype = nodetreetype
    self.clipboard = clipboard
    self._registered = {}
    self._registered_names = []    
  def register(self, node, nodetype):      
    global nodecounter    
    assert nodetype in ("worker", "spydermap", "drone"), nodetype
    assert node not in self._registered    
    nodeclassname = "HiveNodeProxy%4d" % nodecounter 
    nodedict = dict (
      bl_idname = nodeclassname,
      bl_label = node.split()[-1],      
      nodename = node,
      nodetype = nodetype,
      nodetreetype = self.nodetreetype,
      nodecanvas = self.nodecanvas,
    )
    nodecounter += 1
    
    print("REGISTER", nodeclassname)
    nodeclass = type(nodeclassname, (Node, HiveNodeProxy,), nodedict)
    bpy.utils.register_class(nodeclass)
    self._registered[node] = nodeclass
  def unregister(self):    
    for name in self._registered_names:
      nodeclass = self._registered.pop(name)
      bpy.utils.unregister_class(nodeclass)
