"""
Creates a controller around an model object
Use this controller to manipulate the object

Do NOT use the controller to retrieve the value of the object

Files are recognized as a special case: their values are set with a string,
  but actually a File object is constructed

"""
from __future__ import print_function
from functools import partial
from .model import model_basic


class controller_basic:
  def __init__(self,model):
    self._model = model
    self._get = self._model.get
    self._set = self._model.set

class controller(object):
  def __init__(self, form, model, has_parent = False, model_is_func = False):
    self._form = None
    self._model = None
    self._properties = None    
    if form is None: 
      assert has_parent == True
      return
    import Spyder, spyder
    assert isinstance(form, spyder.core.spyderform)
    if hasattr(form, "type") and form.type == "none":
      assert has_parent == True
      return
    self._form = form
    if model_is_func:
      self._model = model
    else:
      self._model = lambda: model
    membernames = form.get_membernames()
    self._setter = self._vsetter
    if membernames is None and self._form.arraycount == 0: 
      assert has_parent == True
      is_file = False 
      if hasattr(form, "type") and form.type == "file": 
        is_file = True
        self._spyderfiletype = Spyder.File
      if hasattr(form, "spydertype") and form.spydertype is not None:
        spydertype = form.spydertype
      else:
        spydertype = getattr(Spyder, form.typename)        
      try:
        if issubclass(spydertype, Spyder.File):
          is_file = True
          self._spyderfiletype = spydertype
      except TypeError:
        pass
            
      if is_file:
        self._setter = self._filesetter
      return
    if self._form.arraycount > 0:
      assert hasattr(form, "length") and form.length is not None
      membernames = [str(v) for v in range(form.length)]
      getfunc = lambda v: form.__getitem__(int(v))
      #getfunc2 = lambda v: self._model().__getitem__(int(v))
      getfunc2 = self._submodel
      modelfunc = True
    else:
      getfunc = partial(getattr, form)
      getfunc2 = self._model().children.__getitem__
      modelfunc = False
      
    self._properties = {}
    for membername in membernames:
      try:
        if modelfunc == False:
          f = controller(getfunc(membername), getfunc2(membername), True)
        else:
          f = controller(getfunc(membername), partial(getfunc2,int(membername)), True, True)      
        if f._form is not None: self._properties[membername] = f
      except Exception as e:
        print("Spyder member:", membername)
        raise
      
  def _submodel(self, item):
    #print("SUB!", item)
    return self._model()[item]
    
  def _vsetter(self, v, copy=False):    
    self._model().set(v, copy)
    
  def _filesetter(self, filename,copy=False):
    self._model().set(self._spyderfiletype(filename, self._form.file, "r"), copy)

  def __getitem__(self, key):
    #print("GET!", key)    
    if not isinstance(key, int): raise TypeError
    if not self._form.arraycount > 0: raise TypeError
    if key < 0: key = self._form.length + key
    if key >= self._form.length: raise IndexError
    if not key >= 0: raise IndexError
    try:
      ret = self._model()[key]
    except (KeyError,IndexError):
      raise IndexError(key)      
    if isinstance(ret, model_basic):
      #return ret.get()
      #return ret
      #print("/GET! BASIC", key)    
      return controller_basic(ret)
    else:
      try:
        ret = self._properties[str(key)]
       # print("/GET! prop", key)            
        return ret
      except KeyError:
        raise IndexError(key)

  def __setitem__(self, key, value):    
    if not isinstance(key, int): raise TypeError
    if not self._form.arraycount > 0: raise TypeError
    if key < 0: key = self._form.length + key
    if key >= self._form.length: raise IndexError
    if not key >= 0: raise IndexError
    kkey = str(key)
    #if kkey not in self._properties: raise KeyError(kkey)
    return self._properties[kkey]._setter(value)

  def __getattr__(self, attr):
    #print("GETA!", attr)
    try:
      ret = self._model().children[attr]
    except KeyError:
      raise AttributeError(attr)      
    if isinstance(ret, model_basic):
      ret = controller_basic(ret)
      #print("/GETA! basic", attr)
      return ret
    else:
      try:
        ret = self._properties[attr]
        #print("/GETA! prop", attr)
        return ret
      except KeyError:
        raise AttributeError(attr)

  def __setattr__(self, attr, value):
    if attr.startswith("_"): 
      return object.__setattr__(self, attr, value)
    try:
      ret = self._properties[attr]._setter(value)
      return ret
    except KeyError:
      raise AttributeError(attr)
      
  def __len__(self): return len(self._model())
  def dict(self):
    if self._model().value is None: return None
    if isinstance(self._model(), model_basic): return self._model().value
    return self._model().value.dict()
  def __eq__(self, other):
    if other is None: return self._model().dif() is None
    return object.__eq__(self, other)    
  def __ne__(self, other):
    return not self.__eq__(other)
  def _exc(self):
    return self._model().exc
  def _get(self):
    return self._model().get()
  def _dif(self):
    return self._model().dif()
  def _set(self, v):
    return self._setter(v)
  def _copy(self, v):
    return self._setter(v,copy=True)

  def __iter__(self): return controller_iterator(self)

class controller_iterator(object):
  def __init__(self, parent):
    self.parent = parent
    self.counter = 0
  def __iter__(self): 
    return self
  def next(self):
    try:
      ret = self.parent[self.counter]
      self.counter += 1
      if isinstance(ret, controller):
        if ret._model().dif() == None: raise IndexError      
      elif ret is None: 
        raise IndexError
      return ret
    except IndexError:
      raise StopIteration
