"""
fill:
Fills the attributes of obj with values from form
Form values are determined in the same priority as for GUI generation
(.value => .default from parents => .default => None)
"""

def _get_value(form, default):
  if hasattr(form, "value"):
    value = form.value
  elif default is not None:
    value = default
  elif hasattr(form, "default"):
    value = form.default
  else:
    value = None
  return value

def _primary(form):
  if form.arraycount == 0 and form.get_membernames() is None: return True
  if hasattr(form, "type") and form.type not in ("required", "none", None): return True
  return False
   
def fill(obj, form, formdefault=None):
  #if hasattr(form, "type") and form.type == "none": return
  if form.arraycount > 0:
    assert hasattr(form, "length") and form.length >= 0
    for n in range(form.length):
      if hasattr(form[n], "type") and form[n].type == "none": continue      
      memdefault = None
      if formdefault is not None: 
        if n < len(formdefault):
          memdefault = formdefault[n]
      memvalue = _get_value(form[n], memdefault)
      if _primary(form[n]):
        #print form.typename
        obj[n] = memvalue
      else:      
        fill(obj[n], form[n], memvalue)
  else:
    assert form.get_membernames() is not None
    for membername in form.get_membernames():
      memdefault = None
      if formdefault is not None: memdefault = getattr(formdefault, membername)
      mform = getattr(form, membername)
      if hasattr(mform, "type") and  mform.type == "none": continue
      memvalue = _get_value(mform, memdefault)      
      if _primary(mform):
        setattr(obj, membername, memvalue)
      else:      
        mobj = getattr(obj, membername)
        #print membername
        fill(mobj, mform, memvalue)
    
   
   
