# Copyright 2007-2011, Sjoerd de Vries
# This file is part of the Spyder module: "core" 
# For licensing information, see LICENSE.txt 

import os, functools
import spyder

def generate_class(typename,parentnames, source, members, deleted_members, block):
  reservednames = ("Spyder", "Exception", "Type", "Delete", "None", "True", "False")
  if (not typename[0].isupper()) or typename.endswith("Error") or typename.endswith("Array") or typename in reservednames:
    raise Exception("compile error: illegal typename for type %s" % typename)
  if block != None: raise Exception
  if len(parentnames):
     s = '\nimport functools\nclass %s(%s):\n  \"\"\"%s\"\"\"\n' % (typename, ",".join(parentnames), source.replace('"""', '\\\"\\\"\\\"'))
  else: s ='\nimport functools\nclass %s(Object):\n  \"\"\"%s\"\"\"\n' % (typename, source.replace('"""', '\\\"\\\"\\\"'))
  s += """
  __multimethods = {} #Auto-generated by Spyder
  """
  if len(parentnames):
    for parentname in parentnames: 
      s += """
  if "_%s__multimethods" in %s.__dict__: __multimethods.update(%s._%s__multimethods) #Auto-generated by Spyder
  """ % (parentname, parentname,parentname,parentname)
  
  s += """
  multimethod = functools.partial(spyder.core.__multimethod__,__multimethods) #Auto-generated by Spyder
  """
  s += """  
  @staticmethod
  def typename():
    \"\"\"Auto-generated by Spyder:
     module core
     file class.py
     function generate_class
    Return the class name of the current object\"\"\"
    return "%s"
""" % typename
  s += """  def cast(self, othertype):
    \"\"\"Auto-generated by Spyder:
     module core
     file class.py
     function generate_class
    Return an object of the type "othertype" initialized from this object\"\"\"
    if type(othertype) == type(int): return othertype(self)    
    if hasattr(othertype, "wrapclass"): return othertype.wrapclass(self)
    return globals()[othertype](self)
"""
  return s, None

def generate_endclass(typename, parentnames, source, members, deleted_members, block, endclass):
  requiredmembers, defaultmembers, optionalmembers, args, allargs = spyder.core.parse_members(typename,members,None, spyder.safe_eval)
  s = ""
  if block != None:
    s += block
  s += "spyder.__types__[\"%s\"] = %s\n" % (typename, typename)    
  s += "%s._requiredmembers = %s\n" % (typename, repr(requiredmembers))
  s += "%s._defaultmembers = %s\n" % (typename, repr(defaultmembers))
  s += "%s.__form__()\n" % typename
  
  s += """
%s.empty = functools.partial(spyder.__constructor, 
  "constructor_empty",
  %s,
  "constructor_fromany",
)
%s.fromlist = functools.partial(spyder.__constructor, 
  "constructor_fromlist",
  %s,
  "constructor_fromany",
)
%s.fromdict = functools.partial(spyder.__constructor, 
  "constructor_fromdict",
  %s,
  "constructor_fromany",
)
""" % (typename,typename,typename,typename,typename,typename)
  
  #generate Array
  s += "for n in range(spyder.arraymax):\n" 
  s += "  classname = \"%s\" + (n+1) * \"Array\"\n" % typename
  s += "  prevclassname = \"%s\" + n * \"Array\"\n" % typename
  s += "  d = dict(typename=\"%s\", classname=classname, prevclassname=prevclassname)\n" % typename
  s += "  defcode = 'class %s(spyder.core.spyderlist):\\n' % classname\n"
  s += "  defcode += '  @staticmethod\\n  def typename(): return \"%s\"\\n' % classname\n"
  cast = """  def cast(self, othertype):
    if type(othertype) == type: return othertype(self)
    return globals()[othertype](self)
""" 
  s += "  defcode += \"\"\"%s\"\"\"\n" % cast
  for b in endclass:
    if endclass[b] != None:
      s += "  defcode += \"\"\"%s\\n\"\"\" %% d\n" % endclass[b].replace('"""','\\"\\"\\"')
  s += "  defcode += \"spyder.core.defineconverter(\\\"%(classname)s\\\",\\\"%(prevclassname)s\\\",\\\"SPLIT\\\")\\n\" % d\n"          
  s += "  defcode += \"spyder.core.defineconverter(\\\"%(prevclassname)s\\\",\\\"%(classname)s\\\",\\\"CAST\\\")\\n\" % d\n"  
  s += "  exec(defcode)\n"
  s += "  spyder.__types__[classname] = globals()[classname]\n"
  s += """
  arrayclass = spyder.__types__[classname]
  %s.empty = functools.partial(spyder.__constructor, 
   "constructor_empty",
   %s,
   "constructor_fromany",
  )
  %s.fromlist = functools.partial(spyder.__constructor, 
   "constructor_fromlist",
   %s,
   "constructor_fromany",
  )
  %s.fromdict = functools.partial(spyder.__constructor, 
   "constructor_fromdict",
   %s,
   "constructor_fromany",
  )
  %s.__form__()
""" % ("arrayclass","arrayclass","arrayclass","arrayclass","arrayclass","arrayclass","arrayclass")

  return s

spyder.defineunimethod("__class", generate_class)
spyder.defineunimethod("__endclass", generate_endclass)
