# Copyright 2007-2011, Sjoerd de Vries
# This file is part of the Spyder module: "core" 
# For licensing information, see LICENSE.txt 


import os, sys
oldpath = list(sys.path)
sys.path.append(os.getcwd())
from parse_members import parse_members
import spyderthread
sys.path = oldpath

class ValidationError(AssertionError): 
  pass


def generate_init(typename, parentnames, source, members, deleted_members, block): 
  requiredmembers, defaultmembers, optionalmembers, args, allargs = spyder.core.parse_members(typename,members,block, spyder.safe_eval)  
  #ignore the block as it contains optional member list
  s = ""
  constructors ="""
  __constructor__ = "constructor_fromany" #Auto-generated by Spyder
  __defaultconstructor__ = __constructor__ #Auto-generated by Spyder
  def constructor_empty(self):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    Empty constructor, all members must be assigned manually afterwards\"\"\"    
  def constructor_fromlist(self, *args, **args2):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    List constructor\"\"\"        
    a = self.__unpack__(*args[0])
    self.__construct__ (*a)
  def constructor_fromdict(self, dic):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    Dict constructor\"\"\"       
    if isinstance(dic, dict) and "_ghostfile" in dic:   
      f = spyder.core.newnamespace()
      f.filename = dic["_ghostfile"][1:-1]
      self._ghostparam = [f, f]
      def getobj(obj):
        return obj                
      self.__dict__ = {"__print__":self._ghost__print__, "__str__":self.__str__, "__repr__": self.__repr__}      
      for k in dir(type(self)):
        if k == "__getattr__" or k in self.__dict__: continue
        self.__dict__[k] = functools.partial(self.__getattr__, k)   
      return                  
    a = self.__copydict__(dic)
    self.__construct__ (*a)              
"""  
  if not len(parentnames) or len(members) > 0:
    s += constructors
  

    typenames = set()
    allmembers = list(requiredmembers)
    allmembers += defaultmembers
    for mem in allmembers:
      typenames.add(mem[0])
    constructor = "%s.__dict__[%s.__constructor__]" % (typename, typename)
  else:
    constructor = "%s.constructor_fromany" % (typename)
  s += """ 
  def __init__(self,*args, **args2):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    Init function, calls the general constructor (fromany by default)\"\"\"            
    self.Annotation = {}
    for funcname in self.__multimethods:
      func, arg = self.__multimethods[funcname]
      self.__dict__[funcname] = functools.partial(func, **({arg:self}))    
    return %s(self, *args, **args2)
  def constructor_fromany(self,*args, **args2):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    General constructor for the current class
    This function tries five specialized constructors in this order:
    - Copy constructor
    - Dict constructor
    - Value/Keyword constructor
    - Parsing constructor
    - List constructor
    For more information, see the Spyder documentation\"\"\"
""" % (constructor)
  if len(members) == 0:
    for parentname in parentnames:
      s+= "    %s.__init__(self,*args,**args2)\n" % parentname
    s+= "    self.validate()\n"
  else:
    s += """
    raise_exc = None
    try:
      exc = "An error has occurred in the construction of a %(t)s object"
      exc += "\\nNumber of arguments: value %%d, keyword %%d\\n" %% (len(args), len(args2))
      try:
        construct = -1
        try:
          if len(args) > 1 or len(args2) > 0: raise IndexError
          a = self.__copy__(args[0]) #func(func(a,b,c))
        except AssertionError as e: 
          e.__context__ = None
          raise
        except Exception as e:
          if len(args) > 1 or len(args2) > 0: raise IndexError
          exc += "\\n*** (%(t)s) Copy constructor ***\\n"+ spyder.exception()
          e.__context__ = None
          try:
            a = self.__copydict__(args[0]) #func({a:1,b:2,c:3})
          except Exception as e:          
            exc += "\\n*** (%(t)s) Dict constructor ***\\n"+ spyder.exception()
            e.__context__ = None
            raise
          else:
            construct = 2
        else:
          construct = 1              
        self.__construct__ (*a)
      except AssertionError as e: 
        e.__context__ = None
        raise
      except Exception as e:
        e.__context__ = None
        if not (len(args) > 1 or len(args2) > 0):
          if construct == 1: exc += "\\n*** Copy constructor (%(t)s) ***\\n" + spyder.exception()
          if construct == 2: exc += "\\n*** Dict constructor (%(t)s) ***\\n" + spyder.exception()
        try:
          if len(args) == 1 and len(args2) == 0 and isinstance(args[0], str):
            if args[0].startswith(\"%(t)s\") and args[0][len(\"%(t)s\"):].lstrip()[0] == "(":
              raise TypeError
          a = self.__unpack__(*args, **args2) #func(a,b,c)
          self.__construct__ (*a) ##
        except AssertionError as e: 
          e.__context__ = None
          raise
        except Exception as e:
          exc += "\\n*** (%(t)s) Value/Keyword constructor ***\\n" + spyder.exception()
          e.__context__ = None
          if len(args) > 1 or len(args2) > 0 : raise          
          try:
            a = self.__parse__(args[0]) #func("[a,b,c]")
            if a == None: return
            self.__construct__ (*a)
          except AssertionError as e: 
            e.__context__ = None
            raise
          except Exception as e:
            exc += "\\n*** (%(t)s) Parsing constructor ***\\n" + spyder.exception()
            e.__context__ = None
            try:
              a = self.__unpack__(*args[0]) #func([a,b,c])
              self.__construct__ (*a)
            except Exception as e:
              exc += "\\n*** (%(t)s) List constructor ***\\n" + spyder.exception()
              e.__context__ = None
              raise
      exc = "" 
      self.validate()
    except AssertionError: raise
    except Exception as err:
      if len(str(err).strip().strip('\\n')) == 0: raise
      else: raise_exc = type(err)(exc+'\\n'+str(err))
    if raise_exc: raise raise_exc  
""" % dict(t = typename)
  arraycode ="""
  __constructor__ = "constructor_fromany"
  __defaultconstructor__ = __constructor__
  def constructor_empty(self):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    Empty constructor, all members must be assigned manually afterwards\"\"\"
  def constructor_fromlist(self, *a):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    List constructor\"\"\"    
    list.__init__(self, *a)
    self.validate()
  def constructor_fromdict(self, *a):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    Dict constructor\"\"\"    
    list.__init__(self, *a)
    self.validate()      
"""    
  arraycode += """  
  def __init__(self,*a, **args2):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    Init function, calls the general constructor (fromany by default)\"\"\"
    if len(args2.keys()): raise Exception          
    self.Annotation = {}
    return %(classname)s.__dict__[%(classname)s.__constructor__](self, *a)  
  def constructor_fromany(self, *a):
    \"\"\"Auto-generated by Spyder:
     module core
     file init.py
     function generate_init
    General constructor for arrays\"\"\"    
    if len(a) == 0:
      spyder.core.spyderlist.__init__(self)
      return
    try:
      if len(a) == 1 and isinstance(a[0],str): raise Exception
      self.__copy__(*a)
    except AssertionError: raise
    except:
      try:
        if isinstance(a,str): raise Exception
        self.__copy__(a)      
      except AssertionError: raise
      except:
        if len(a) > 1: raise
        self.__parse__(a[0])
"""
  #if len(members) == 0: return None,arraycode    
  return s,arraycode

spyder.defineunimethod("__init__", generate_init)

