# Copyright 2007-2011, Sjoerd de Vries
# This file is part of the Spyder module: "core" 
# For licensing information, see LICENSE.txt 

def generate_validate(typename, parentnames, source, members, deleted_members, block):
  requiredmembers, defaultmembers, optionalmembers, args, allargs = spyder.core.parse_members(typename,members,None, spyder.safe_eval)  
  if block != None: raise Exception
  s = """  def validate(self):
    \"\"\"Auto-generated by Spyder:
     module core
     file validate.py
     function generate_validate
    Validates the current object
    This is automatically called after construction
    It is NOT automatically called after editing the object\"\"\"
"""
  s += "    try:\n      pass\n"
  for m in requiredmembers:
    s += "      self.%s = %s(self.%s)\n" % (m[1],m[0],m[1])
  for m in defaultmembers:
    if m[1] not in optionalmembers:
      s += "      self.%s = %s(self.%s)\n" % (m[1],m[0],m[1])
  s += "      self.__validate__()\n"
  s += """    except Exception as inst: 
      if isinstance(inst, AssertionError):
        raise 
      else:
        #TODO; import sys
        #tb = sys.exc_info()[2]      
        #e = spyder.core.ValidationError
        #raise e, e(str(inst)), tb #TODO
        raise spyder.core.ValidationError(str(inst))
"""
  arraycode = "  def validate(self): self.__validate__()\n"
  if len(members) == 0 and not len(parentnames): 
    return None,arraycode    
  return s,arraycode

def generate_intern_validate(typename, parentnames, source, members, deleted_members, block):
  requiredmembers, defaultmembers, optionalmembers, args, allargs = spyder.core.parse_members(typename,members,None, spyder.safe_eval)  
  s = """  def __validate__(self):
    \"\"\"Auto-generated by Spyder:
     module core
     file validate.py
     function generate_intern_validate
    Private object validator, for internal use only\"\"\"
"""
  if block == None:
    s += "    pass\n"
  else:
    ss = ""
    for m in requiredmembers:
      ss += "    %s = self.%s\n" % (m[1],m[1])
    for m in defaultmembers:
      ss += "    %s = self.%s\n" % (m[1],m[1])      
    spaces = -1
    for l in block.split('\n'):
      if len(l.strip()) == 0: continue
      if spaces == -1:
        spaces = len(l) - len(l.lstrip())
        s += ss
      s += "    %s\n" % l[spaces:]    
    if spaces == -1: s += "    pass\n"

  arraycode = """  def __arrayvalidate__(self):
    if self.__constructor__ == self.__defaultconstructor__:
      def reconstruct(classtype, var):
        return classtype(var)
    else:
       def reconstruct(classtype, var):         
         classtype.__constructor__ = self.__constructor__
         try:
           ret = classtype(var)
         finally:
           classtype.__constructor__ = self.__defaultconstructor__
         return ret               
    for v in range(len(self)):
      self[v] = reconstruct(%(prevclassname)s,self[v])
  def __validate__(self):
    try:
      self.__arrayvalidate__()
    except Exception as inst:
      if isinstance(inst, AssertionError):
        raise 
      else: raise spyder.core.ValidationError(str(inst))
"""
  if len(members) == 0 and not len(parentnames): return None,arraycode    
  return s,arraycode

spyder.defineunimethod("__validate__", generate_validate)
spyder.defineunimethod("validate", generate_intern_validate)
#NOTE that they are swapped, to send "validate" block to "__validate__" function,
# and make constant "validate" function not to accept blocks!


