import spyder
import Spyder
from Spyder import Object, File, String, Integer, Bool, Float
from functools import partial

class qid(str):
  def __new__(self, value=None):
    if value is not None: value = value.replace("-","_")
    return str.__new__(self, value)
    
class qstr(str):
  def __repr__(self):
    return '"' + str(self) + '"'

class qexpr(str):
  def qstr(self): 
    return qstr(self)
  def __repr__(self):
    return str(self)

import copy
class qdict(object):
  def __init__(self, _type=None, **members):
    self._type = _type
    self._items = []
    for member in members:
      if member == "items":
        for key, value in members["items"]: 
          self._items.append((key,value))
      else:
        value = members[member]
        self._items.append((member,value))  
  def __deepcopy__(self, memo):
    ret = qdict(self._type)
    for key, value in self._items:
      v = copy.deepcopy(value, memo)
      ret._items.append((key, v))
    return ret
  def __getattr__(self, key):
    return self[key]
  def __getitem__(self, key):
    if isinstance(key, int):
      return self._items[key]
    else:
      items = [v[1] for v in self._items if v[0] == key]
      if len(items) == 0: raise KeyError(key)
      elif len(items) > 1: raise KeyError("Duplicate key: %s, %d" % (key, len(items)))
      else: return items[0]
  def __setattr__(self, key, value):
    if key.startswith("_"): 
      object.__setattr__(self, key, value)
      return
    self[key] = value
  def __setitem__(self, key, value):
    if isinstance(key, int):
      self._items[key] = value
    else:
      items = [vnr for vnr,v in enumerate(self._items) if v[0] == key]
      if len(items) == 0: self._items.append((key,value))  
      elif len(items) > 1: raise KeyError("Duplicate key: %s, %d" % (key, len(items)))
      else: self._items[item[0]] = value
  def append(self, item):
    self._items.append((None,item))  
  def merge(self, item):
    assert isinstance(item, qdict)
    if item._type is None: 
      self._items += item._items
    else:
      self._items.append((None,item))  
  def items(self):
    return self._items
  def keys(self):
    return (v[0] for v in self._items)
  def values(self):
    return (v[0] for v in self._items)
  def __iter__(self):
    return self.keys()
  def __repr__(self, spaces=0):
    spacestr = " " * spaces
    ret = "" #first line will be indented by parent
    if self._type is not None: ret += self._type + " "
    ret += "{\n"
    spaces2 = spaces + tabstep
    spacestr2 = " " * spaces2
    for key, value in self._items:
      if key is not None: 
        ret += "%s%s: %s\n" % (spacestr2, key, repr(value))
      else:
        assert isinstance(value,qdict), value
        ret += spacestr2 + value.__repr__(spaces2)
    ret += "%s}\n" % spacestr
    return ret
  
###############################################################################
# SETTINGS
###############################################################################

form_head = """import QtQuick 1.1
import QtDesktop 0.1
"""
form_tail = ""


form_init = qdict("Rectangle", width=400)
tabstep = 2

#if "wrappings" is between these values, the generated QML depends on "wrappings"
# therefore, if wrappings changes (due to grouping), the QML must then be regenerated
wrappings_regen_min = 1000
wrappings_regen_max = 1000

###############################################################################

_generators = {}
_group_generators = {}
_ele_generators = {}
_phase_generators = {}
_tokens = ("outerwrap", "midwrap", "head", "innerwrap", "group") 
_phases = ("outer", "pre", "post", "mid", "inner")
_allowedformtypes = ("required", "none", "text", "password", "number", "slider", "spin", "checkbox", "option", "radio", "file")  

def set_generator(token, complexity, callback):
  """
  See the HTML generator
  """
  assert isinstance(complexity, int) and complexity >= 0
  _generators[token, complexity] = callback

def set_group_generator(grouptype, complexity, callback):
  assert isinstance(complexity, int) and complexity >= 0
  _group_generators[grouptype, complexity] = callback

def set_element_generator(element, callback):
  _ele_generators[element] = callback

def set_phase_generator(element, complexity, phase, callback):
  """
  See htmlform
  """
  assert phase in _phases, phase
  _phase_generators[element, complexity] = phase, callback
  
from . import generators
generators.init()

###########################################################################


from .resolve import resolve
      
def gen(qml, generator, args):
  if generator is None: return
  generator(qml, *args)

def process_other_tokens(phase, qml, props, complexity, data):
  if "_othertokens" not in props: return
  name, id, objs, forms, wrappings = data
  tokens = props["_othertokens"]
  
  for token, args in tokens:
    if (token, complexity) not in _phase_generators:
      raise ValueError("Unknown token %s, complexity %s" % (token, complexity))
    tokphase, callback = _phase_generators[token, complexity]
    if tokphase != phase: continue
    cb_args = (name, id, objs, forms, wrappings, args)
    if callback is not None: callback(qml, *cb_args)

def _write_qml(objs, forms, name=None,prename="",wrappings=0):  

  obj, form = objs[0], forms[0]

  if hasattr(form, "value"):
    defaultvalue = form.value
  elif obj is not None:
    defaultvalue = obj
  elif hasattr(form, "default"):
    defaultvalue = form.default
  else:
    defaultvalue = None
 
  complexity = 0
  mqmls = []
  mqmldata = [] #stores some values in case we have to regenerate QML
  
  n0, n = None, None
  if form.typename is None:
    typnam = "None" + "Array" * form.arraycount
  else:
    typnam = form.typename + "Array" * form.arraycount
  if name is not None:
    n0 = prename + name
    n = prename.replace(".", "-") + name
   
  curr_membernames = None
  if form.arraycount > 0:
    is_array = True
    is_none = False   
    if hasattr(form, "type"):
      if form.type not in ("none", "required"): is_array = False
      if form.type == "none": is_none = True
    if is_array:
      if is_none is False:
        if not hasattr(form, "length"):
          raise TypeError("Cannot generate QML for \"%s\"(%s): member is an Array, but no length has been specified" % (n0, typnam))
        if not isinstance(form.length, int) or form.length <= 0:
          raise TypeError("Cannot generate QML for \"%s\"(%s): member is an Array, but its length is not a positive integer" % (n0, typnam))
        curr_membernames = []
        curr_members = []
        curr_objs = []
        for nr in range(form.length):
          curr_membernames.append(str(nr))
          curr_members.append(form[nr])
          mobj = None
          if defaultvalue is not None and len(defaultvalue) > nr:
            mobj = defaultvalue[nr]
          curr_objs.append(mobj)  
  elif form.get_membernames():
    curr_membernames = form.get_membernames()
    curr_members = []
    curr_objs = []
    for membername in curr_membernames:
      curr_members.append(getattr(form, membername))
      mobj = None
      if defaultvalue is not None: 
        mobj = getattr(defaultvalue, membername)
      curr_objs.append(mobj)
      
  mwrappings = wrappings + 1
  if curr_membernames is not None:    
    if hasattr(form, "memberorder"):
      mems = []
      for mnr, m in enumerate(form.memberorder):
        assert m not in form.memberorder[mnr+1:]
        assert m in curr_membernames, m        
        mems.append(curr_membernames.index(m))
      for mnr, membername in enumerate(curr_membernames):        
        if membername not in form.memberorder:
          mems.append(mnr)
      curr_membernames = [curr_membernames[i] for i in mems]      
      curr_members = [curr_members[i] for i in mems]
      curr_objs = [curr_objs[i] for i in mems]
    mcomplexities = []
    for membername, mform, mobj in zip(curr_membernames, curr_members, curr_objs):
      newprename = prename
      if name is not None: 
        newprename += name + "."
      mobjs = [mobj, defaultvalue] + objs[1:]
      mforms = [mform] + forms
      mqml, mcomplexity = _write_qml(mobjs, mforms, membername,newprename,mwrappings)
      mqmls.append(mqml)
      mqmldata.append((mobjs, mforms, newprename))
      if hasattr(mform, "group") and mform.group is not None:
        mcomplexity += 1
      mcomplexities.append(mcomplexity)
    complexity = max(mcomplexities) 
    if form.arraycount == 0 or complexity == 0: complexity +=  1  
 
  for token in _tokens:
    assert token,complexity in _generators  
  gen_group = _generators["group", complexity]
  gen_innerwrap = _generators["innerwrap", complexity]
  if name is None:
    gen_head = None
    gen_outerwrap = None
    gen_midwrap = None
  else:
    gen_head = _generators["head", complexity]    
    gen_outerwrap = _generators["outerwrap", complexity]  
    gen_midwrap = _generators["midwrap", complexity]
  qml = qdict()
  while 1:
    if name is not None:
      if form.is_default:
        if not hasattr(form, "type") and defaultvalue is None:
          raise TypeError("Cannot generate QML for \"%s\"(%s): member is optional, but no default value and no type (required, none or other) is specified" % (n0, typnam))
        if hasattr(form, "type") and form.type == "none": break
      else:
        if hasattr(form, "type") and form.type == "none":
          raise TypeError("Cannot generate QML for \"%s\"(%s): member is not optional, but type is specified as \"none\"" % (n0, typnam))

    tokdata = (name, n, objs, forms, wrappings)

    process_other_tokens("outer", qml, form._props, complexity, tokdata)

    gen_args = (name, n, objs, forms, wrappings)
    if gen_outerwrap is not None: gen_outerwrap(qml, *gen_args)    

    process_other_tokens("pre", qml, form._props, complexity, tokdata)

    if gen_midwrap is not None: gen_midwrap(qml, *gen_args)    

    process_other_tokens("mid", qml, form._props, complexity, tokdata)
    
    name2 = name
    if hasattr(form, "name"): name2 = form.name
    if gen_head is not None: 
      gen_head(qml, name, n, objs, forms, wrappings, name2)

    process_other_tokens("post", qml, form._props, complexity, tokdata)

    if gen_innerwrap is not None: gen_innerwrap(qml, *gen_args)    

    process_other_tokens("inner", qml, form._props, complexity, tokdata)
    
    curr_wrappings = mwrappings
    if complexity > 0:
      
      done_membernames = set()
      groupindex = 0
      resolve(
       qml,
       curr_wrappings, 
       done_membernames,
       mqmls,
       mqmldata,
       mwrappings,
       curr_membernames,
       curr_members,
       mcomplexities,
       groupindex,
       name,
       n,
       objs,
       forms,
       wrappings,
       complexity,
       gen_group,
       toplevel = True
      )
    else: #complexity 0
      typ = getattr(Spyder, form.typename)
      formtype = "text"
      if issubclass(typ, File):
        if form.arraycount > 0:
          raise TypeError("Cannot generate QML for \"%s\"(%s): array of files" % (n0, typnam))
        if not hasattr(form, "file"):
          raise TypeError("Cannot generate QML for \"%s\"(%s): a file type (file attribute) must be specified" % (n0, typnam))
        if hasattr(form, "options"):
          raise TypeError("Cannot generate QML for \"%s\"(%s): file elements cannot have options" % (n0, typnam))
        formtype = "file"
      elif issubclass(typ, Integer) or issubclass(typ, Float):
        formtype = "spin"
        
      elif issubclass(typ, Bool):
        formtype = "checkbox"
       
      if hasattr(form, "options"):
        assert formtype == "text"
        formtype = "option"
        options = form.options      
        optiontitles = options
        if hasattr(form, "optiontitles"):
          optiontitles = form.optiontitles
          assert len(options) == len(optiontitles), (n, len(options), len(optiontitles))
     
      if hasattr(form, "type"):
        if form.type not in ("required", "none"):
          formtype = form.type
        
      assert formtype in _allowedformtypes or formtype in _ele_generators, formtype

      formsubtype = None
      if hasattr(form, "subtype"):
        formsubtype = form.subtype
      if formsubtype is None:
        if issubclass(typ, Integer):
          formsubtype = "int"
        elif issubclass(typ, Float):
          formsubtype = "float"

      expand = True
      if hasattr(form, "expand"): expand = form.expand
                  
      if formtype in _ele_generators:
        gener = _ele_generators[formtype]
        gener(name, n, qml, objs, forms, wrappings, defaultvalue) 
      else: #standard types
        
        if formtype == "option":
          assert hasattr(form, "options"), n
          if defaultvalue is not None:
            assert defaultvalue in options, (n, defaultvalue, options)          
          
          qml._type = "SelectionDialog"
          id = qstr("_widget-%s" % n)
          qml.id = qid(id)
          qml.objectName = id
          qml.titleText = name2
          model = qdict("ListModel")
          qml.model = model
          for opt, optitle in zip(options, optiontitles):
            le = qdict("ListElement",name=opt,text=optitle)
            model.append(le)
          if defaultvalue is not None:
            for onr,opt in enumerate(options):
              if opt == defaultvalue:
                qml.selectedIndex = onr
                break
        elif formtype == "radio":
          assert hasattr(form, "options"), n
          if defaultvalue is not None:
            assert defaultvalue in options, (n, defaultvalue, options)          

          eqml = qdict("Rectangle")
          eqml["border.color"] = qstr("black")
          qml.append(eqml)
          
          cg = qdict("CheckableGroup")
          id = qstr("_group-%s" % n)
          cg.id = qid(id)
          cg.objectName = id
          eqml.append(cg)
          label = qdict("Text")
          id =  qstr("_head-%s" % n)
          label.id = qid(id)
          label.objectName = id
          label.text = qstr(name2)
          eqml.append(label)
          column = qdict("Column")
          eqml.append(column)
          
          for onr in range(len(options)):
            opt, optitle = options[onr], optiontitles[onr]
            oid = qstr("_widget-%s-%d" % (n, onr))
            rb = qdict("RadioButton", text=optitle,platformExclusiveGroup=label.id)
            rb.id = qid(oid)
            rb.objectName = oid
            if defaultvalue is not None and opt == defaultvalue:
              rb.checked = qexpr("true")
            column.append(rb)
        elif formtype == "file":
          raise Exception #TODO
          
          xml[0].append(t.tabstr + '<object class="GtkFileChooserButton" id="_widget-%s">' % n)
          t.incr()
          xml[0].append(t.tabstr + '<property name="visible">True</property>')          
          xml[0].append(t.tabstr + '<property name="can_focus">True</property>')
          xml[0].append(t.tabstr + '<property name="hexpand">%s</property>' % expand)
          xml[0].append(t.tabstr + '<property name="vexpand">%s</property>' % expand)
          xml[0].append(t.tabstr + '<property name="orientation">vertical</property>')
          if defaultvalue is not None:
            if isinstance(defaultvalue, File):
              defaultstr = defaultvalue.name
            elif isinstance(defaultvalue, str):
              defaultstr = defaultvalue
            xml[0].append(t.tabstr + '<property name="active">%s</property>' % defaultstr)
          if hasattr(form, "folder") and form.folder == True:
            xml[0].append(t.tabstr + '<property name="action">select-folder</property>')
          t.decr()
          xml[0].append(t.tabstr + '</object>')
          ###
        elif formtype == "checkbox":
          qml._type = "CheckBox"
          id = qstr("_widget-%s" % n)
          qml.id = qid(id)
          qml.objectName = id
          qml.text = qstr(name2)
          if defaultvalue is not None:
            if defaultvalue:
              qml.checked = qstr("true")
            else:
              qml.checked = qstr("false")
        elif formtype == "spin":
          step = 1
          if hasattr(form, "step"): step = form.step
          elif hasattr(form, "range"): 
            step = form.range/10.0
          if formsubtype == "int": step = int(step+0.999)

          digits = 0
          if hasattr(form, "digits"): digits = form.digits
          elif formsubtype == "float":
            digits = 1
            if step < 1 and step > 0:
              while 10**-digits > step: digits += 1

          #spinbutton
          ####
          eqml = qdict("Rectangle")
          eqml["border.color"] = qstr("black")
          qml.append(eqml)
          
          label = qdict("Text")
          id =  qstr("_head-%s" % n)
          label.id = qid(id)
          label.objectName = id
          label.text = qstr(name2)          
          eqml.append(label)
          
          spin = qdict("SpinBox")
          qml.append(spin)
          id = qstr("_widget-%s" % n)
          spin.id = qid(id)
          spin.objectName = id
                    
          if hasattr(form, "length"):
            #TODO
            pass
          #TODO: digits

          minimum = -999999
          if hasattr(form, "min"): minimum = form.min
          spin.minimumValue = minimum          
          maximum = 999999
          if hasattr(form, "max"): maximum = form.max
          spin.maximumValue = minimum          
          if defaultvalue is not None:
            spin.value = defaultvalue

          spin.singleStep = step

        elif formtype == "text" or formtype == "password":
          eqml = qdict("Rectangle", height = 30, width = 200)
          eqml["border.color"] = qstr("black")
          qml.append(eqml)

          label = qdict("Text")
          id = qstr("_head-%s" % n)
          label.id = qid(id)
          label.objectName = id
          label.text = qstr(name2)          
          label.width, label.height = 200,30
          eqml.append(label)

          eqml = qdict("Rectangle", height = 30, width = 200)
          eqml["border.color"] = qstr("black")
          qml.append(eqml)          
          inp = qdict("TextInput")
          inp.width, inp.height = 200,30
          eqml.append(inp)
          if formtype == "password":
            inp.echoMode = qexpr("TextInput.Password")

          defaultstr = ""
          if defaultvalue is not None:
            defaultstr = qstr(defaultvalue)
          if defaultvalue is not None:
            inp.text = defaultstr
          if hasattr(form, "length"):
            inp.maximumLength = length
        else:        
          raise Exception(formtype)


      ### ENDIF standard types    
    ### ENDIF complexity == 0
      
    break

  return qml, complexity


def qml(obj=None, form=None):
  """
  obj can be: A Spyder class, a Spyder object, or None
    if obj is None, then form must not be None
  form is a spyderform object
    if form is None, it is extracted from the Spyder obj
  """
    
  if form is None:
    if isinstance(obj, Object):
      form = obj._form()
    else:
      ok = False
      try: 
        if issubclass(obj, Object):
          form = obj._form() 
          obj = None
          ok = True
      except TypeError:
        pass
      if not ok: raise TypeError(obj)
  else:
    try: 
      if issubclass(obj, Object): 
        obj = None
    except TypeError:
      pass

  assert isinstance(form, spyder.core.spyderform)
  
  qml = copy.deepcopy(form_init)
  qml0, complexity = _write_qml([obj], [form])
  qml.merge(qml0)
  ret = form_head + "\n"
  ret += str(qml)
  return ret
    
  

